﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.SimulationCore {
  [StorableClass]
  public abstract class DiscreteEventSimulation<TModel> : Simulation, IDiscreteEventSimulation<TModel> where TModel : class, IModel {
    [Storable]
    public IAction<TModel> InitialAction { get; protected set; }
    [Storable]
    public IEnumerable<IActivity<TModel>> Activities { get; protected set; }
    [Storable]
    public IEnumerable<IReporter<TModel>> Reporters { get; protected set; }
    [Storable]
    public TModel Model { get; protected set; }

    protected CompositeEventQueue<TModel> CompositeEventQueue { get; private set; }

    protected IFixedValueParameter<DoubleValue> MaximumTimeParameter {
      get { return (IFixedValueParameter<DoubleValue>)Parameters["MaximumTime"]; }
    }

    public double MaximumTime {
      get { return MaximumTimeParameter.Value.Value; }
      set { MaximumTimeParameter.Value.Value = value; }
    }

    protected DiscreteEventSimulation(DiscreteEventSimulation<TModel> original, Cloner cloner)
      : base(original, cloner) {
      InitialAction = cloner.Clone(original.InitialAction);
      Activities = original.Activities.Select(cloner.Clone).ToList();
      Reporters = original.Reporters.Select(cloner.Clone).ToList();
      Model = cloner.Clone(original.Model);
    }

    protected DiscreteEventSimulation() {
      Parameters.Add(new FixedValueParameter<DoubleValue>("MaximumTime", "The simulation stops after reaching this time.", new DoubleValue(1000)));
    }

    protected override void Run(CancellationToken cancellationToken) {
      CompositeEventQueue = new CompositeEventQueue<TModel>(Activities.Select(x => x.EventQueue));

      var actionMonitor = new Dictionary<Guid, List<IActivity<TModel>>>();
      foreach (var activity in Activities) {
        foreach (var actionId in activity.MonitoredActionIds) {
          if (!actionMonitor.ContainsKey(actionId)) actionMonitor[actionId] = new List<IActivity<TModel>>();
          actionMonitor[actionId].Add(activity);
        }
      }

      if (InitialAction != null) {
        InitialAction.Execute(Model);
        CallActivities(actionMonitor, InitialAction);
      }

      while (!CompositeEventQueue.Empty && Model.CurrentTime < MaximumTime) {
        var @event = CompositeEventQueue.Pop();
        var nextEvent = CompositeEventQueue.Empty ? null : CompositeEventQueue.Peek();
        var action = @event.Action;

        action.Execute(Model);
        CallActivities(actionMonitor, action);

        if (@event.Time > Model.CurrentTime && (nextEvent == null || @event.Time < nextEvent.Time)) {
          var timeUpdateAction = new ModelTimeUpdateAction<TModel>(null, @event.Time);
          timeUpdateAction.Execute(Model);

          if (actionMonitor.ContainsKey(timeUpdateAction.ActionId))
            foreach (var activity in actionMonitor[timeUpdateAction.ActionId])
              activity.ManageEvents(Model, timeUpdateAction);

          foreach (var reporter in Reporters)
            reporter.UpdateReporting(Model, Results);
        }
        cancellationToken.ThrowIfCancellationRequested();
      };
    }

    private void CallActivities(Dictionary<Guid, List<IActivity<TModel>>> actionMonitor, IAction<TModel> action) {
      if (actionMonitor.ContainsKey(action.ActionId)) {
        foreach (var activity in actionMonitor[action.ActionId])
          activity.ManageEvents(Model, action);
      }
    }
  }
}
